type = 'MviewFisheyeCalibrator'
work_dir = './temp'
chessboard_width = 6
chessboard_height = 7
chessboard_square_size = 100
calibrate_intrinsic = False
calibrate_distortion = False
calibrate_extrinsic = True
